open BsMocha.Mocha;
open BsJsverify.Verify.Arbitrary;
open BsJsverify.Verify.Property;
let (<.) = Function.Infix.(<.);

/* TODO: move this and other implementations to bs-jsverify */
let arb_belt_set_int: (~min: int, ~max: int) => arbitrary(Belt.Set.Int.t) =
  (~min, ~max) =>
    smap(
      Belt.Set.Int.fromArray,
      Belt.Set.Int.toArray,
      ~newShow=
        l =>
          Js.Json.stringifyAny(Belt.Set.Int.toArray(l))
          |> Js.Option.getWithDefault(""),
      arb_array(arb_int(min, max)),
    );

describe("Belt.Set.Int", () => {
  /* Intentionally restricted range to avoid arithmetic overflows as much as possible */
  let arb_int' =
    arb_int(
      Js.Math.pow_int(~base=Int.Bounded.bottom, ~exp=1 / 4),
      Js.Math.pow_int(~base=Int.Bounded.top, ~exp=1 / 4),
    );

  let arb_belt_set_int' =
    arb_belt_set_int(
      ~min=Js.Math.pow_int(~base=Int.Bounded.bottom, ~exp=1 / 4),
      ~max=Js.Math.pow_int(~base=Int.Bounded.top, ~exp=1 / 4),
    );

  describe("Conjunctive", () => {
    describe("Medial Magma", () => {
      module V = Verify.Medial_Magma(Belt_SetInt.Conjunctive.Medial_Magma);
      property4(
        "should satisfy bicommutativity",
        arb_belt_set_int',
        arb_belt_set_int',
        arb_belt_set_int',
        arb_belt_set_int',
        V.bicommutativity,
      );
    });
    describe("Semigroup", () => {
      module V = Verify.Semigroup(Belt_SetInt.Conjunctive.Semigroup);
      property3(
        "should satisfy associativity",
        arb_belt_set_int',
        arb_belt_set_int',
        arb_belt_set_int',
        V.associativity,
      );
    });
    describe("Monoid", () => {
      module V = Verify.Monoid(Belt_SetInt.Conjunctive.Monoid);
      property1("should satisfy identity", arb_belt_set_int', V.identity);
    });
  });

  describe("Eq", () => {
    module V = Verify.Eq(Belt_SetInt.Eq);
    property1("should satisfy reflexivity", arb_belt_set_int', V.reflexivity);
    property2("should satisfy symmetry", arb_belt_set_int', arb_belt_set_int', V.symmetry);
    property3(
      "should satisfy transitivity",
      arb_belt_set_int',
      arb_belt_set_int',
      arb_belt_set_int',
      V.transitivity,
    );
  });

  describe("Ord", () => {
    module V = Verify.Ord(Belt_SetInt.Ord);
    property1("should satisfy reflexivity", arb_belt_set_int', V.reflexivity);
    property2(
      "should satisfy antisymmetry",
      arb_belt_set_int',
      arb_belt_set_int',
      V.antisymmetry,
    );
    property3(
      "should satisfy transitivity",
      arb_belt_set_int',
      arb_belt_set_int',
      arb_belt_set_int',
      V.transitivity,
    );
  });

  describe("Join_Semilattice", () => {
    module V = Verify.Join_Semilattice(Belt_SetInt.Join_Semilattice);
    property3(
      "should satisfy associativity",
      arb_belt_set_int',
      arb_belt_set_int',
      arb_belt_set_int',
      V.associativity,
    );
    property2(
      "should satisfy commutativity",
      arb_belt_set_int',
      arb_belt_set_int',
      V.commutativity,
    );
    property1("should satisfy idempotency", arb_belt_set_int', V.idempotency);
  });
});
